unit doublePolyhedronUnit;

interface

{
Much of this code is adapted from Chapter 7 of
O'Rourke, Joseph. 1998. "Computational Geometry in C"
(Second Edition). Cambridge University Press 376 p.
}

uses Classes, SysUtils, SolidGeom;

const
  X = 0;
  Y = 1;
  Z = 2;

var
  Epsilon : extended;

type
  TPointi = array[0..DIM-1] of integer;
    // integer point //adapted from "Computational Geometry in C"

  TInPolyhedronResult = (ipVertex, ipEdge, ipFace, ipInterior, ipExterior);
  TSegPlaneIntersectResult = (spiWithin, spiQOnPlane, spiROnPlane, spiOffPlane,
    spiIntersects);
  TInTriangleResult = (itVertex, itEdge, itInterior, itExterior);
  TSegTriangleIntersectionResult = (stiVertex, stiEdge, stiFace, stiMisses,
    stiVertexOnSegment, stiEdgeOnSegment, stiFaceOnSegment, stiDegenerate);
  TBoxTestResult = (btOutside, btUnknown);

  TSegmentObject = class (TObject)
    StartLoc, EndLoc : TPointd;
    function Length : extended;
    function TwoDIntersection(ASegment : TSegmentObject;
      var Intersection : TPointd) : boolean;
  end;

  TSegmentList = Class(TObject)
    private
      FList : TList;
      function Getitems(Index: integer): TSegmentObject;
    public
      function Count : integer;
      Constructor Create;
      Destructor Destroy; override;
      property items[Index : integer] : TSegmentObject read Getitems ; default;
      function IndexOf(ASegment : TSegmentObject) : integer;
      function Add(ASegment: TSegmentObject): Integer;
  end;

  TPolyhedron = class (TObject)
    private
      F : integer; //number of faces
      V : integer; // number of verticies.
      function Inbox(q : TPointd) : boolean;
        //adapted from "Computational Geometry in C"
      procedure RandomRay(var ray : tPointd );
        //adapted from "Computational Geometry in C"
      function BoxTest (  n : integer; a, b : tPointd ) : TBoxTestResult;
        //adapted from "Computational Geometry in C"
      procedure AddVec(q : tPointd; var ray: tPointd );
        //adapted from "Computational Geometry in C"
      procedure NormalVec(  a,  b,  c: tPointd; var  N : tPointd);
        //adapted from "Computational Geometry in C"
      function Dot(  a: tPointd; b: tPointd ) : extended;
        //adapted from "Computational Geometry in C"
      function PlaneCoeff(T : tPointi; var N : tPointd ; var D : extended )
        : integer; //adapted from "Computational Geometry in C"
      procedure SubVec(  a,  b : tPointd; var c : tPointd );
        //adapted from "Computational Geometry in C"
      function SegPlaneInt(T: TPointi; q, r : tPointd; var p : tPointd;
        var m : integer) : TSegPlaneIntersectResult;
        //adapted from "Computational Geometry in C"
      function AreaSign(  a, b, c : tPointd ) : integer;
        //adapted from "Computational Geometry in C"
      function InTri2D(  Tp : array of tPointd;  pp :tPointd )
        : TInTriangleResult;
        //adapted from "Computational Geometry in C"
      function InTri3D(  T : tPointi;  m : integer; p : tPointd )
        : TInTriangleResult;
        //adapted from "Computational Geometry in C"
      function InPlane(  T : tPointi;  m: integer;  q,  r : tPointd;
        p : tPointd) : TSegTriangleIntersectionResult;
        //adapted from "Computational Geometry in C"
      function VolumeSign(  a,  b,  c,  d : tPointd ) : integer;
        //adapted from "Computational Geometry in C"
      function SegTriCross(  T: tPointi;  q,  r :tPointd )
        : TSegTriangleIntersectionResult;
        //adapted from "Computational Geometry in C"
      function SegTriInt(T: tPointi; q, r : tPointd; var p : tPointd  )
        : TSegTriangleIntersectionResult;
      function FacesSame(FaceIndex1, FaceIndex2: Integer) : boolean;
        //adapted from "Computational Geometry in C"
    public
      Vertices : array of TPointd;
        //adapted from "Computational Geometry in C"
      Faces : array of TPointi;
        //adapted from "Computational Geometry in C"
      Box : array of array[0..1] of TPointd;
        //adapted from "Computational Geometry in C"
      bmin, bmax : TPointd;
      radius : extended;
      constructor Create(VertexCount, FaceCount: integer);
      function InPolyHedron( q : TPointd) : TInPolyhedronResult;
        //adapted from "Computational Geometry in C"
      function ComputeBox: extended;
        //adapted from "Computational Geometry in C"
      function SegmentIntersection(Const ASegment : TSegmentObject) : TSegmentList;
      Procedure GetProps(Var Vol, Area : extended);
      function PolygonIntersect(APolygon : TXYZPlane) : TXYZPlaneList;
      function VertexIndex(Avertex : TPointd) : Integer;
      Procedure EliminateInternalFaces(LastFace, LastVertex : integer);
  end;

function SortPointObjects(Item1, Item2: Pointer): Integer;
function WrongEndsMatch(FirstSegment, SecondSegment : TSegmentObject)
  : boolean;
function EndsMatch(FirstSegment, SecondSegment : TSegmentObject) : boolean;

implementation

uses SolidUnit, RealListUnit, IntListUnit, VirtualMeshFunctions;


function SortPointObjects(Item1, Item2: Pointer): Integer;
{
SortPointObjects sorts TPointObject's in order by their X, Y, and Z
coordinates.
}
var
  Point1, Point2 : TPointObject;
begin
  Point1 := Item1;
  Point2 := Item2;
  if Point1.Location[X] < Point2.Location[X] then
  begin
    result := -1;
  end
  else if Point1.Location[X] > Point2.Location[X] then
  begin
    result := 1;
  end
  else if Point1.Location[Y] < Point2.Location[Y] then
  begin
    result := -1;
  end
  else if Point1.Location[Y] > Point2.Location[Y] then
  begin
    result := 1;
  end
  else if Point1.Location[Z] < Point2.Location[Z] then
  begin
    result := -1;
  end
  else if Point1.Location[Z] > Point2.Location[Z] then
  begin
    result := 1;
  end
  else
  begin
    result := 0;
  end;
end;

function TPolyhedron.Inbox(q : TPointd) : boolean;
{
Inbox tests whether q is inside the smallest box with sides parallel
to the principle coordinate axes in which the polygon will fit.

Bmin, and bmax must already have been assigned before calling Inbox.

}
begin
  result := ( ( bmin[X] <= q[X] ) and ( q[X] <= bmax[X] ) and
              ( bmin[Y] <= q[Y] ) and ( q[Y] <= bmax[Y] ) and
              ( bmin[Z] <= q[Z] ) and ( q[Z] <= bmax[Z] ) )
end;

procedure TPolyhedron.RandomRay(var ray : tPointd );
{
   Generate a random point on a sphere of radius 1.
   the sphere is sliced at z, and a random point at angle t
   generated on the circle of intersection.

   Radius must already have been set before calling RandomRay.
}
var
  x_l, y_l, z_l, w, t: extended;
begin

  z_l :=  random;
  t := Pi * random;
  w := sqrt( 1 - Sqr(z_l) );
  x_l := w * cos( t );
  y_l := w * sin( t );

  ray[X] := radius * x_l;
  ray[Y] := radius * y_l;
  ray[Z] := radius * z_l;

end;

function TPolyhedron.BoxTest (  n : integer; a, b : tPointd ) : TBoxTestResult;
{
Test whether the line segment a,b intersects the box around face n

Box must have been set before calling BoxTest.
}
var
  i : integer; //* Coordinate index */
  w : extended;
begin
  for i := 0 to DIM-1 do
  begin
    w := Box[ n ,0,i]; //* min: lower left */
    if ( (a[i] < w) and (b[i] < w) ) then
    begin
      result := btOutside;
      Exit;
    end;
    w := Box[ n ][1][i]; //* max: upper right */
    if ( (a[i] > w) and (b[i] > w) ) then
    begin
      result := btOutside;
      Exit;
    end;
  end;
  result := btUnknown;

end;

procedure TPolyhedron.AddVec(q : tPointd; var ray: tPointd );
{
Add two vectors.
}
var
  i : integer;
begin
  for i := 0 to DIM-1 do
  begin
    ray[i] := q[i] + ray[i]
  end;
end;

procedure TPolyhedron.NormalVec(  a,  b,  c: tPointd; var  N : tPointd);
{
Calculate the normal N to the plane defined by a, by, c.
}
begin
    N[X] := ( c[Z] - a[Z] ) * ( b[Y] - a[Y] ) -
           ( b[Z] - a[Z] ) * ( c[Y] - a[Y] );
    N[Y] := ( b[Z] - a[Z] ) * ( c[X] - a[X] ) -
           ( b[X] - a[X] ) * ( c[Z] - a[Z] );
    N[Z] := ( b[X] - a[X] ) * ( c[Y] - a[Y] ) -
           ( b[Y] - a[Y] ) * ( c[X] - a[X] );
end;

function TPolyhedron.Dot(  a: tPointd; b: tPointd ) : extended;
{
Calculate the dot product of a and b.
}
var
  i : integer;
begin
  result := 0.0;

  for i := 0 to DIM-1 do
  begin
    result := result + a[i] * b[i];
  end;
end;

function TPolyhedron.PlaneCoeff(T : tPointi; var N : tPointd; var D : extended)
  : integer;
{
Compute the normal (N) to Face T and the dot product of N and the first
point in T. Also determine the largest component of N.
}
var
  i : integer;
  t_l : extended; //* Temp storage */
  biggest : extended; //* Largest component of normal vector. */
  m : integer; //* Index of largest component. */
begin
  biggest := 0;
  m := 0;

    NormalVec( Vertices[T[0]], Vertices[T[1]], Vertices[T[2]], N );
    D := Dot( Vertices[T[0]], N );

    //* Find the largest component of N. */

    for i := 0 to DIM -1 do
    begin
      t_l := Abs( N[i] );
      if ( t_l > biggest ) then
      begin
        biggest := t_l;
        m := i;
      end;
    end;

    result := m;

end;

procedure TPolyhedron.SubVec(  a,  b : tPointd; var c : tPointd );
{
Subtract two vectors.
}
var
  i : integer;
begin
  for i := 0 to DIM- 1 do
  begin
    c[i] := a[i] - b[i];
  end;
end;

function TPolyhedron.SegPlaneInt(T: TPointi; q, r : tPointd; var p : tPointd;
  var m : integer) : TSegPlaneIntersectResult;
{
Determine the point of intersection (p) between the face T and the the line
segment defined by q and r. Also determine the largest component of the Normal
to T.
}
var
  N : tPointd;
  D : extended;
  rq : tPointd;
  num, denom, t_l : extended;
  i : integer;
begin

    for i := 0 to DIM-1 do
    begin
      N[i] := 0;
    end;
    D:=0;
    m := PlaneCoeff( T, N, D );
    num := D - Dot( q, N );
    SubVec( r, q, rq );
    denom := Dot( rq, N );

    if ( denom = 0.0 ) then  //* Segment is parallel to plane. */
    begin
       if ( num = 0.0 ) then  //* q is on plane. */
       begin
         result := spiWithin;
         Exit;
       end
       else
       begin
         result := spiOffPlane;
         Exit;
       end;
    end
    else
    begin
       t_l := num / denom;
    end;

    for i := 0 to DIM-1 do
    begin
      p[i] := q[i] + t_l * ( r[i] - q[i] );
    end;

    if ( (0.0 < t_l) and (t_l < 1.0) ) then
    begin
      result := spiIntersects;
    end
    else if ( num = 0.0 ) then  //* t_l == 0 */
    begin
      result := spiQOnPlane
    end
    else if ( num = denom ) then //* t_l == 1 */
    begin
      result := spiROnPlane;
    end
    else
    begin
      result := spiOffPlane;
    end;
end;

function TPolyhedron.AreaSign(  a, b, c : tPointd ) : integer;
var
  area2 : extended;
begin

    area2 := ( b[0] - a[0] ) * ( c[1] - a[1] ) -
            ( c[0] - a[0] ) * ( b[1] - a[1] );

    if      ( area2 >  epsilon ) then
    begin
      result := 1;
    end
    else if ( area2 < -epsilon ) then
    begin
      result := -1;
    end
    else
    begin
      result := 0;
    end;
end;

function TPolyhedron.InTri2D(  Tp : array of tPointd;  pp :tPointd )
  : TInTriangleResult;
{
Determine whether pp is in Tp. Only the first two members of each tPointd
are used.
}
var
  area0, area1, area2 : integer;
begin

   //* compute three AreaSign() values for pp
   // with respect to each edge of the face in 2D */
   area0 := AreaSign( pp, Tp[0], Tp[1] );
   area1 := AreaSign( pp, Tp[1], Tp[2] );
   area2 := AreaSign( pp, Tp[2], Tp[0] );

   if ( ( area0 = 0 ) and ( area1 > 0 ) and ( area2 > 0 ) or
        ( area1 = 0 ) and ( area0 > 0 ) and ( area2 > 0 ) or
        ( area2 = 0 ) and ( area0 > 0 ) and ( area1 > 0 ) ) then
   begin
     result := itEdge;
   end

   else if ( ( area0 = 0 ) and ( area1 < 0 ) and ( area2 < 0 ) or
        ( area1 = 0 ) and ( area0 < 0 ) and ( area2 < 0 ) or
        ( area2 = 0 ) and ( area0 < 0 ) and ( area1 < 0 ) ) then
   begin
     result := itEdge;
   end

   else if ( ( area0 >  0 ) and ( area1 > 0 ) and ( area2 > 0 ) or
        ( area0 <  0 ) and ( area1 < 0 ) and ( area2 < 0 ) ) then
   begin
     result := itInterior;
   end

   else if ( ( area0 = 0 ) and ( area1 = 0 ) and ( area2 = 0 ) ) then
   begin
     Assert(False);
   end

   else if ( ( area0 = 0 ) and ( area1 = 0 ) or
        ( area0 = 0 ) and ( area2 = 0 ) or
        ( area1 = 0 ) and ( area2 = 0 ) ) then
   begin
     result := itVertex;
   end

   else
   begin
     result := itExterior;
   end;
end;

function TPolyhedron.InTri3D(  T : tPointi;  m : integer; p : tPointd )
  : TInTriangleResult;
{
determine whether p lies in T.
}
var
  i : integer; //* Index for X,Y,Z           */
  j : integer; //* Index for X,Y             */
  k : integer; //* Index for triangle vertex */
  pp : tPointd; //* projected p */
  Tp : array [0..2] of tPointd; //* projected T: three new vertices */
begin

   //* Project out coordinate m in both p and the triangular face */
   j := 0;
   for i := 0 to DIM-1 do
   begin
     if ( i <> m ) then    //* skip largest coordinate */
     begin
       pp[j] := p[i];
       for k := 0 to 2 do
       begin
         Tp[k][j] := Vertices[T[k]][i];
       end;
       Inc(j);
     end;
   end;
   result := InTri2D( Tp, pp ) ;
end;

function TPolyhedron.InPlane(  T : tPointi;  m: integer;  q,  r : tPointd;
  p : tPointd) : TSegTriangleIntersectionResult;
begin
    //* NOT IMPLEMENTED */
    result := stiDegenerate;
//    return 'p';
end;

function TPolyhedron.VolumeSign(  a,  b,  c,  d : tPointd ) : integer;
var
  vol : extended ;
  ax, ay, az, bx, by, bz, cx, cy, cz, dx, dy, dz : extended ;
  bxdx, bydy, bzdz, cxdx, cydy, czdz : extended ;
begin
   ax := a[X];
   ay := a[Y];
   az := a[Z];
   bx := b[X];
   by := b[Y];
   bz := b[Z];
   cx := c[X];
   cy := c[Y];
   cz := c[Z];
   dx := d[X];
   dy := d[Y];
   dz := d[Z];

   bxdx:=bx-dx;
   bydy:=by-dy;
   bzdz:=bz-dz;
   cxdx:=cx-dx;
   cydy:=cy-dy;
   czdz:=cz-dz;
   vol :=  (az-dz) * (bxdx*cydy - bydy*cxdx)
         + (ay-dy) * (bzdz*cxdx - bxdx*czdz)
         + (ax-dx) * (bydy*czdz - bzdz*cydy);

   // This may not work for floating point values.
   if      ( vol > Epsilon ) then
   begin
     result := 1;
   end
   else if ( vol < -Epsilon )then
   begin
     result := -1;
   end
   else
   begin
     result := 0;
   end;
end;

function TPolyhedron.SegTriCross(  T: tPointi;  q,  r :tPointd )
  : TSegTriangleIntersectionResult;
var
  vol0, vol1, vol2 : integer;
begin
  if VertexIndex(q)>-1 then
  begin
    result := stiVertex;
  end;


  vol0 := VolumeSign( q, Vertices[ T[0] ], Vertices[ T[1] ], r );
  vol1 := VolumeSign( q, Vertices[ T[1] ], Vertices[ T[2] ], r );
  vol2 := VolumeSign( q, Vertices[ T[2] ], Vertices[ T[0] ], r );

  //* Same sign: segment intersects interior of triangle. */
  if ( ( ( vol0 > 0 ) and ( vol1 > 0 ) and ( vol2 > 0 ) ) or
       ( ( vol0 < 0 ) and ( vol1 < 0 ) and ( vol2 < 0 ) ) ) then
  begin
     result := stiFaceOnSegment;
  end

  //* Opposite sign: no intersection between segment and triangle */
  else if ( ( ( vol0 > 0 ) or ( vol1 > 0 ) or ( vol2 > 0 ) ) and
       ( ( vol0 < 0 ) or ( vol1 < 0 ) or ( vol2 < 0 ) ) ) then
  begin
     result := stiMisses
  end


  else if ( ( vol0 = 0 ) and ( vol1 = 0 ) and ( vol2 = 0 ) ) then
  begin
    Assert(False);
  end

  //* Two zeros: segment intersects vertex. */
  else if ( ( ( vol0 = 0 ) and ( vol1 = 0 ) ) or
            ( ( vol0 = 0 ) and ( vol2 = 0 ) ) or
            ( ( vol1 = 0 ) and ( vol2 = 0 ) ) ) then
  begin
     result := stiVertex
  end

  //*/ One zero: segment intersects edge. */
  else if ( ( vol0 = 0 ) or ( vol1 = 0 ) or ( vol2 = 0 ) ) then
  begin
     result := stiEdge
  end


  else
  begin
    Assert(False);
  end;
end;

function TPolyhedron.SegTriInt(T: tPointi; q, r : tPointd; var p : tPointd  )
  : TSegTriangleIntersectionResult;
var
  code : TSegPlaneIntersectResult;
  m : integer;
begin
    m := -1;

    code := SegPlaneInt( T, q, r, p, m );

    case code of
    spiOffPlane:
      begin
        result := stiMisses;
      end;
    spiQOnPlane :
      begin
        case InTri3D( T, m, q ) of
          itVertex :
          begin
            result := stiVertex;
          end;
          itEdge :
          begin
            result := stiEdge;
          end;
          itInterior :
          begin
            result := stiFace;
          end;
          itExterior :
          begin
            result := stiMisses;
          end;
        end;
      end;
    spiROnPlane :
      begin
        case InTri3D( T, m, r ) of
          itVertex :
          begin
            result := stiVertex;
          end;
          itEdge :
          begin
            result := stiEdge;
          end;
          itInterior :
          begin
            result := stiFace;
          end;
          itExterior :
          begin
            result := stiMisses;
          end;
        end;
      end;
    spiWithin :
      begin
        result := InPlane( T, m, q, r, p );
      end;
    spiIntersects :
      begin
        result := SegTriCross( T, q, r );
      end;
    else
      begin
        Assert(False);
      end
    end;

end;

function TPolyhedron.InPolyHedron( q : TPointd) : TInPolyhedronResult;
var
  r : TPointd; // ray endpoint
  p : TPointd; // intersection point; not used
  f_l: integer;
  crossings : integer;
  code : TSegTriangleIntersectionResult;
  VertexIndex, DimIndex : integer;
  Same : boolean;
  label Loop;
begin
  if not InBox(q ) then
  begin
    result := ipExterior;
    Exit;
  end;

  for VertexIndex := 0 to V-1 do
  begin
    Same := True;
    for DimIndex := 0 to DIM-1 do
    begin
      Same := Same and
        NearlyTheSame(q[DimIndex], Vertices[VertexIndex,DimIndex],Epsilon);
      if not Same then
      begin
        break;
      end;
    end;
    if Same then
    begin
      result := ipVertex;
      Exit;
    end;
  end;


  crossings := 0;
  Loop:
  While True do
  begin
    crossings := 0;
    RandomRay( r  );
    AddVec( q, r );
    for f_l := 0 to F-1 do //* Begin check each face */
    begin
      if ( BoxTest( f_l, q, r ) = btOutside) then
      begin
        code := stiMisses;
      end
      else
      begin
        code := SegTriInt( Faces[f_l], q, r, p )
      end;
      case code of
      stiDegenerate, stiVertexOnSegment, stiEdgeOnSegment:
        begin
          Goto Loop;
          // degenerate, start over
        end;
      stiFaceOnSegment:
        begin
          Inc(crossings);
        end;
      stiVertex:
        begin
          result := ipVertex;
          Exit;
        end;
      stiEdge:
        begin
          result := ipEdge;
          Exit;
        end;
      stiFace:
        begin
          result :=  ipFace;
          Exit;
        end;
      stiMisses:
        begin
          // do nothing.
        end;
      else
        begin
          Assert(False);
        end;
        end;

    end;
    break;

  end;
  if Odd(crossings) then
  begin
    result := ipInterior;
  end
  else
  begin
    result := ipExterior;
  end;

end;

constructor TPolyhedron.Create(VertexCount, FaceCount: integer);
var
  Index, InnerIndex : integer;
begin
  inherited Create;
  V := VertexCount;
  SetLength(Vertices,VertexCount);
  F := FaceCount;
  SetLength(Faces,FaceCount);
  SetLength(Box,FaceCount);
  for Index := 0 to V-1 do
  begin
    for InnerIndex := 0 to DIM-1 do
    begin
      Vertices[Index,InnerIndex] := 0;
    end;
  end;
  for Index := 0 to F-1 do
  begin
    for InnerIndex := 0 to DIM-1 do
    begin
      Faces[Index,InnerIndex] := 0;
    end;
  end;
end;

function TPolyhedron.ComputeBox : extended;
var
  i, j: integer;
  FaceIndex : integer;
  AFace : TPointi;
  AVertex : TPointd;
  VertexIndex, InnerVertexIndex : integer;
begin

  for i := 0 to DIM-1 do
  begin
    bmin[i] := Vertices[0][i];
    bmax[i] := Vertices[0][i];
  end;

  for i := 0 to V-1 do
  begin
    for j := 0 to DIM-1 do
    begin
      if( Vertices[i][j] < bmin[j] ) then
      begin
	bmin[j] := Vertices[i][j];
      end;
      if( Vertices[i][j] > bmax[j] ) then
      begin
	bmax[j] := Vertices[i][j];
      end;
    end;
  end;

  radius := sqrt( Sqr(bmax[X] - bmin[X]) +
                 Sqr(bmax[Y] - bmin[Y]) +
                 Sqr(bmax[Z] - bmin[Z]) ) *2;

  for FaceIndex := 0 to F-1 do
  begin
    AFace := Faces[FaceIndex];
    AVertex := Vertices[AFace[0]];
    Box[FaceIndex,0] := AVertex;
    Box[FaceIndex,1] := AVertex;
    for VertexIndex := 1 to DIM -1 do
    begin
      AVertex := Vertices[AFace[VertexIndex]];
      for InnerVertexIndex := 0 to DIM -1  do
      begin
        if AVertex[InnerVertexIndex] < Box[FaceIndex,0,InnerVertexIndex] then
        begin
          Box[FaceIndex,0,InnerVertexIndex] := AVertex[InnerVertexIndex];
        end;
        if AVertex[InnerVertexIndex] > Box[FaceIndex,1,InnerVertexIndex] then
        begin
          Box[FaceIndex,1,InnerVertexIndex] := AVertex[InnerVertexIndex];
        end;
      end;

    end;
  end;

  result :=  radius +1;
end;

function TPolyhedron.SegmentIntersection(
  Const ASegment: TSegmentObject): TSegmentList;
var
  VertexList, OrderList : TList;
  StartPoint, EndPoint : TPointObject;
  FaceIndex : integer;
  IntersectionLocation : tPointd;
  code : TSegTriangleIntersectionResult;
  IntermediatePoint : TPointObject;
  FirstPointOK : boolean;
  Index : integer;
  Point1, Point2 : TPointObject;
  NewSegment : TSegmentObject;
  PointIndex : integer;
  Edge, Line : TSegmentObject;
  EdgeList : TList;
  EdgeIndex : integer;
  fraction : double;
begin
  result := TSegmentList.Create;
  VertexList := TList.Create;
  OrderList := TList.Create;
  StartPoint := TPointObject.Create;
  EndPoint := TPointObject.Create;
  try
    OrderList.Capacity := 2;
    StartPoint.Location := ASegment.StartLoc;
    EndPoint.Location := ASegment.EndLoc;
    OrderList.Add(StartPoint);
    OrderList.Add(EndPoint);
    VertexList.Add(StartPoint);
    if InPolyHedron(StartPoint.Location ) <> ipExterior then
    begin
      FirstPointOK := True;
    end
    else
    begin
      FirstPointOK := False;
    end;
    if InPolyHedron(EndPoint.Location ) <> ipExterior then
    begin
      VertexList.Add(EndPoint);
    end;
    for FaceIndex := 0 to F -1 do
    begin
      code := SegTriInt(Faces[FaceIndex], StartPoint.Location,
        EndPoint.Location, IntersectionLocation);
      if code = stiDegenerate then
      begin
        EdgeList := TList.Create;
        Line := TSegmentObject.Create;
        try
          EdgeList.Capacity := DIM;
          Line.StartLoc := StartPoint.Location;
          Line.EndLoc := EndPoint.Location;

          for PointIndex := 0 to DIM -2 do
          begin
            Edge := TSegmentObject.Create;
            Edge.StartLoc := Vertices[Faces[FaceIndex][PointIndex]];
            Edge.EndLoc := Vertices[Faces[FaceIndex][PointIndex+1]];
            EdgeList.Add(Edge);
          end;

          Edge := TSegmentObject.Create;
          Edge.StartLoc := Vertices[Faces[FaceIndex][DIM -1]];
          Edge.EndLoc := Vertices[Faces[FaceIndex][0]];
          EdgeList.Add(Edge);

          for EdgeIndex := 0 to EdgeList.Count -1 do
          begin
            Edge := EdgeList[EdgeIndex];
            if Line.TwoDIntersection(Edge,IntersectionLocation) then
            begin
              fraction := Sqrt(Sqr(IntersectionLocation[X] - Line.StartLoc[X])
                + Sqr(IntersectionLocation[Y] - Line.StartLoc[Y]))/
                  Sqrt(Sqr(Line.StartLoc[X] - Line.EndLoc[X])
                + Sqr(Line.StartLoc[Y] - Line.EndLoc[Y]));

              IntersectionLocation[Z] := Line.StartLoc[Z]
                + fraction*(Line.EndLoc[Z] - Line.StartLoc[Z] );

              IntermediatePoint := TPointObject.Create;
              IntermediatePoint.Location := IntersectionLocation;
              VertexList.Add(IntermediatePoint);
            end;
          end;

        finally
          EdgeList.Free;
          Line.Free;
        end;

      end
      else if (code <> stiMisses) then
      begin
        IntermediatePoint := TPointObject.Create;
        IntermediatePoint.Location := IntersectionLocation;
        VertexList.Add(IntermediatePoint);
      end;
    end;
    VertexList.Sort(SortPointObjects);
    OrderList.Sort(SortPointObjects);
    if OrderList.IndexOf(StartPoint) <> 0 then
    begin
      for Index := 0 to (VertexList.Count -1) div 2 do
      begin
        VertexList.Exchange(Index, VertexList.Count -1 -Index);
      end;
    end;
    if not FirstPointOK then
    begin
      VertexList.Remove(StartPoint);
    end;
    if VertexList.Count > 1 then
    begin
      for Index := 0 to VertexList.Count -2 do
      begin
        Point1 := VertexList[Index];
        Point2 := VertexList[Index+1];
        if not NearlyTheSame(Point1.Location[X], Point2.Location[X],
                 Epsilon) or
           not NearlyTheSame(Point1.Location[Y], Point2.Location[Y],
                 Epsilon) or
           not NearlyTheSame(Point1.Location[Z], Point2.Location[Z],
                 Epsilon) then
        begin
          IntersectionLocation[X] := (Point1.Location[X]
            + Point2.Location[X])/2;
          IntersectionLocation[Y] := (Point1.Location[Y]
            + Point2.Location[Y])/2;
          IntersectionLocation[Z] := (Point1.Location[Z]
            + Point2.Location[Z])/2;
          if InPolyHedron(IntersectionLocation ) <> ipExterior then
          begin
            NewSegment := TSegmentObject.Create;
            NewSegment.StartLoc := Point1.Location;
            NewSegment.EndLoc := Point2.Location;
            result.FList.Add(NewSegment);
          end;
        end;
      end;

    end;

  finally
    if VertexList.IndexOf(StartPoint) < 0 then
    begin
      StartPoint.Free;
    end;
    if VertexList.IndexOf(EndPoint) < 0 then
    begin
      EndPoint.Free;
    end;
    for Index := 0 to VertexList.Count -1 do
    begin
      TPointObject(VertexList[Index]).Free;
    end;
    VertexList.Free;
    OrderList.Free;
  end;

end;


{ TSegmentList }

function TSegmentList.Add(ASegment: TSegmentObject): Integer;
begin
  result := FList.Add(ASegment);
end;

function TSegmentList.Count: integer;
begin
  result := FList.Count;
end;

constructor TSegmentList.Create;
begin
  inherited;
  FList := TList.Create;

end;

destructor TSegmentList.Destroy;
begin
  FList.Free;
  inherited;
end;

function TSegmentList.Getitems(Index: integer): TSegmentObject;
begin
  result := FList[Index];
end;


function TSegmentList.IndexOf(ASegment: TSegmentObject): integer;
begin
  result := FList.IndexOf(ASegment);
end;

procedure TPolyhedron.GetProps(var Vol, Area: extended);
var
  ASolid : TSolid;
  Xs, Ys, Zs : TRealList ;
  FaceLists : TIntListList;
  Index : integer;
  AFaceList : TIntegerList;
  VertexIndex : integer;
begin
  Xs := TRealList.Create;
  Ys := TRealList.Create;
  Zs := TRealList.Create;
  FaceLists := TIntListList.Create;
  try
    Xs.Capacity := V;
    Ys.Capacity := V;
    Zs.Capacity := V;
    FaceLists.Capacity := F;
    begin
      for Index := 0 to V-1 do
      begin
        Xs.Add(Vertices[Index,X]);
        Ys.Add(Vertices[Index,Y]);
        Zs.Add(Vertices[Index,Z]);
      end;
      for Index := 0 to F-1 do
      begin
        AFaceList := TIntegerList.Create;
        FaceLists.Add(AFaceList);
        AFaceList.Add(Index);
        for VertexIndex := DIM -1 downto 0 do
        begin
          AFaceList.Add(Faces[Index,VertexIndex]);
        end;
      end;

      ASolid := TSolid.Create(Xs,Ys,Zs,FaceLists);
      try
        ASolid.GetProps(Vol, Area);
      finally
        ASolid.Free;
      end;
    end
  finally
    begin
      Xs.Free;
      Ys.Free;
      Zs.Free;
      for Index := 0 to FaceLists.Count -1 do
      begin
        TIntegerList(FaceLists[Index]).Free;
      end;
      FaceLists.Free;
    end;
  end;

end;

function WrongEndsMatch(FirstSegment, SecondSegment : TSegmentObject)
  : boolean;
begin
  result := (NearlyTheSame(FirstSegment.EndLoc[0],SecondSegment.EndLoc[0],
               Epsilon))
    and (NearlyTheSame(FirstSegment.EndLoc[1], SecondSegment.EndLoc[1],
           Epsilon))
    and (NearlyTheSame(FirstSegment.EndLoc[2], SecondSegment.EndLoc[2],
           Epsilon));
end;

function EndsMatch(FirstSegment, SecondSegment : TSegmentObject) : boolean;
begin
  result := (NearlyTheSame(FirstSegment.EndLoc[0], SecondSegment.StartLoc[0],
               Epsilon))
    and (NearlyTheSame(FirstSegment.EndLoc[1], SecondSegment.StartLoc[1],
           Epsilon))
    and (NearlyTheSame(FirstSegment.EndLoc[2], SecondSegment.StartLoc[2],
           Epsilon));
end;

function FirstEndsMatch(FirstSegment, SecondSegment : TSegmentObject) : boolean;
begin
  result := (NearlyTheSame(FirstSegment.StartLoc[0], SecondSegment.StartLoc[0],
               Epsilon))
    and (NearlyTheSame(FirstSegment.StartLoc[1], SecondSegment.StartLoc[1],
           Epsilon))
    and (NearlyTheSame(FirstSegment.StartLoc[2], SecondSegment.StartLoc[2],
           Epsilon));
end;

function ReversedEndsMatch(FirstSegment, SecondSegment : TSegmentObject) : boolean;
begin
  result := (NearlyTheSame(FirstSegment.StartLoc[0], SecondSegment.EndLoc[0],
               Epsilon))
    and (NearlyTheSame(FirstSegment.StartLoc[1], SecondSegment.EndLoc[1],
           Epsilon))
    and (NearlyTheSame(FirstSegment.StartLoc[2], SecondSegment.EndLoc[2],
           Epsilon));
end;

function TPolyhedron.PolygonIntersect(APolygon: TXYZPlane): TXYZPlaneList;
  Function GetNextSegment(AList : TList; ASegment : TSegmentObject)
    : TSegmentObject;
  var
    AnotherSegment : TSegmentObject;
    Index : integer;
    ALocation : TPointd;
  begin
    result := nil;
    for Index := 0 to AList.Count -1 do
    begin
      AnotherSegment := AList[Index];
      if (ASegment<>AnotherSegment) then
      begin
        if  EndsMatch(ASegment, AnotherSegment) then
        begin
          result := AnotherSegment;
          break;
        end
        else if WrongEndsMatch(ASegment, AnotherSegment) then
        begin
          ALocation := AnotherSegment.StartLoc;
          AnotherSegment.StartLoc := AnotherSegment.EndLoc;
          AnotherSegment.EndLoc := ALocation;
          result := AnotherSegment;
          break;
        end
      end;
    end;

  end;
var
  Converter : TPointObject;
  ASegment, AnotherSegment : TSegmentObject;
  SegmentList : TSegmentList;
  AList, AnotherList : TList;
  Index : integer;
  InnerIndex : integer;
  AnotherPolygon : TXYZPlane;
  FacePointer : TPointi;
  vertexIndex, FaceIndex : integer;
  Procedure CreatePolygon;
  var
    Index : integer;
    APlane : TXYZPlane;
    NodeIndex, InnerNodeIndex : integer;
  begin
    if AnotherList.Count > 2 then
    begin
      APlane := TXYZPlane.Create(AnotherList.Count);
      result.Add(APlane);
      for Index := 0 to AnotherList.Count-1 do
      begin
        ASegment := AnotherList[Index];
        Converter.Location := ASegment.StartLoc;
        APlane.Outline[Index] := Converter.XYZLocation;
      end;
      Converter.Location := ASegment.EndLoc;
      if ((NearlyTheSame(APlane.Outline[0].X, APlane.Outline[APlane.Count-1].X,
                Epsilon)) and
             (NearlyTheSame(APlane.Outline[0].Y, APlane.Outline[APlane.Count-1].Y,
                Epsilon)) and
             (NearlyTheSame(APlane.Outline[0].Z, APlane.Outline[APlane.Count-1].Z,
                Epsilon)) ) then
      begin
        if APlane.Count = 3 then
        begin
          result.Remove(APlane);
          APlane.Free;
          APlane := nil;
        end
        else
        begin
          APlane.Count := APlane.Count -1;
        end;
        if APlane <> nil then
        begin
          for NodeIndex := APlane.Count -1 downto 1 do
          begin
            if NearlyTheSame(APlane.Outline[NodeIndex].X, APlane.Outline[NodeIndex-1].X, Epsilon) and
               NearlyTheSame(APlane.Outline[NodeIndex].Y, APlane.Outline[NodeIndex-1].Y, Epsilon) and
               NearlyTheSame(APlane.Outline[NodeIndex].Z, APlane.Outline[NodeIndex-1].Z, Epsilon) then
            begin
              if APlane.Count = 3 then
              begin
                result.Remove(APlane);
                APlane.Free;
              end
              else
              begin
                for InnerNodeIndex := NodeIndex to APlane.Count -2 do
                begin
                  APlane.Outline[NodeIndex] := APlane.Outline[NodeIndex+1]
                end;
                APlane.Count := APlane.Count -1;
              end;
            end;
          end;
        end;
      end;
    end;
    for Index := 0 to AnotherList.Count-1 do
    begin
      TSegmentObject(AnotherList[Index]).Free;
    end;
    AnotherList.Clear;
    if AList.Count > 0 then
    begin
      ASegment := AList[0];
      AnotherList.Add(ASegment);
      AList.Remove(ASegment);
    end;
  end;
var
  DimIndex : integer;
  TempSegList : TList;
  TempSegIndex : integer;
  SegmentIndex, InnerSegmentIndex : integer;
begin
  result := TXYZPlaneList.Create;
  AList := TList.Create;
  AnotherList := TList.Create;
  Converter := TPointObject.Create;
  try
    for DimIndex := 0 to DIM -1 do
    begin
      if (APolygon.bmin[DimIndex] > bmax[DimIndex]) or
         (APolygon.bmax[DimIndex] < bmin[DimIndex]) then
      begin
        Exit;
      end;
    end;

    // Get the parts of the outline of the APolygon that are in the
    // polyhedron.
    for Index := 0 to APolygon.Count -2 do
    begin
      ASegment := TSegmentObject.Create;
      try
        Converter.XYZLocation := APolygon.Outline[Index];
        ASegment.StartLoc := Converter.Location;
        Converter.XYZLocation := APolygon.Outline[Index+1];
        ASegment.EndLoc := Converter.Location;
        SegmentList := SegmentIntersection(ASegment);
        try
          for InnerIndex := 0 to SegmentList.Count -1 do
          begin
            AList.Add(SegmentList[InnerIndex]);
          end;
        finally
          SegmentList.Free;
        end;
      finally
        ASegment.Free;
      end;
    end;
    ASegment := TSegmentObject.Create;
    try
      Converter.XYZLocation := APolygon.Outline[APolygon.Count-1];
      ASegment.StartLoc := Converter.Location;
      Converter.XYZLocation := APolygon.Outline[0];
      ASegment.EndLoc := Converter.Location;
      SegmentList := SegmentIntersection(ASegment);
      try
        for InnerIndex := 0 to SegmentList.Count -1 do
        begin
          AList.Add(SegmentList[InnerIndex]);
        end;
      finally
        SegmentList.Free;
      end;
    finally
      ASegment.Free;
    end;

    // Get the intersection between each face of the polyhedron and APolygon
    for FaceIndex := 0 to F-1 do
    begin
      FacePointer := Faces[FaceIndex];
      AnotherPolygon := TXYZPlane.Create(DIM);
      try
        for VertexIndex := 0 to DIM-1 do
        begin
          Converter.Location := Vertices[FacePointer[VertexIndex]];
          AnotherPolygon.Outline[VertexIndex] := Converter.XYZLocation;
        end;
        TempSegList := TList.Create;
        try
          if (APolygon.PlaneIntersection(AnotherPolygon, TempSegList)
            = piOutlinesCross) then
          begin
            if TempSegList.Count > 0 then
            begin
              AList.Capacity := AList.Count + TempSegList.Count;
              for TempSegIndex := 0 to TempSegList.Count -1 do
              begin
                ASegment := TempSegList[TempSegIndex];
                AList.Add(ASegment);
              end;
            end;

          end;
        finally
          TempSegList.Free;
        end;

      finally
        AnotherPolygon.Free;
      end;

    end;

    for SegmentIndex := AList.Count -2 downto 0 do
    begin
      ASegment := AList[SegmentIndex];
      for InnerSegmentIndex := AList.Count -1 downto SegmentIndex + 1 do
      begin
        AnotherSegment := AList[InnerSegmentIndex];
        if (FirstEndsMatch(ASegment, AnotherSegment)
              and WrongEndsMatch(AnotherSegment,ASegment))
           or (ReversedEndsMatch(ASegment, AnotherSegment)
              and ReversedEndsMatch(AnotherSegment, ASegment)) then
        begin
          AnotherSegment.Free;
          AList.Delete(InnerSegmentIndex);
        end;
      end;

    end;


    // arrange the segments.
    if AList.Count > 0 then
    begin
      ASegment := AList[0];
      AnotherList.Add(ASegment);
      AList.Remove(ASegment);
      While AList.Count > 0 do
      begin
        ASegment := GetNextSegment(AList,ASegment);
        if ASegment = nil then
        begin
          CreatePolygon;
        end
        else
        begin
          AnotherList.Add(ASegment);
          AList.Remove(ASegment);
        end;
      end;
    end;
    // create the resulting polygon
    CreatePolygon;
  finally
    Converter.Free;

    for Index := 0 to AList.Count-1 do
    begin
      TSegmentObject(AList[Index]).Free;
    end;
    AList.Free;

    for Index := 0 to AnotherList.Count-1 do
    begin
      TSegmentObject(AnotherList[Index]).Free;
    end;
    AnotherList.Free;
  end;

end;

function TPolyhedron.VertexIndex(Avertex: TPointd): Integer;
var
  Index : integer;
  AnotherVertex : TPointd;
begin
  result := -1;
  for Index := 0 to V-1 do
  begin
    AnotherVertex := Vertices[Index];
    if (Avertex[X] = AnotherVertex[X]) and
       (Avertex[Y] = AnotherVertex[Y]) and
       (Avertex[Z] = AnotherVertex[Z]) then
    begin
      result := Index;
      break;
    end;
  end;
end;

procedure TPolyhedron.EliminateInternalFaces(LastFace,
  LastVertex: integer);
var
  FaceIndex1, FaceIndex2 : integer;
  FaceIndex : integer;
  DuplicateFound : Boolean;
  NodeIndex : integer;
  NodeUsed : array of boolean;
//  InnerNodeIndex : integer;
  EliminatedNodes : integer;
  NodeIndicies : array of integer;
begin
  // eliminate parts of the arrays that weren't used.
  if V <> LastVertex+1 then
  begin
    V := LastVertex+1;
    Setlength(Vertices,V);
  end;

  if F <> LastFace+1 then
  begin
    F := LastFace+1;
    Setlength(Faces,F);
  end;
  // Get rid of duplicate faces because they are internal faces.
  repeat
    DuplicateFound := False;
    for FaceIndex1 := F-2 downto 0 do
    begin
      for FaceIndex2 := F-1 downto FaceIndex1 + 1 do
      begin
        if FacesSame(FaceIndex1, FaceIndex2) then
        begin
          DuplicateFound := True;
          for FaceIndex := FaceIndex1+1 to FaceIndex2 -1 do
          begin
            Faces[FaceIndex-1] := Faces[FaceIndex];
          end;
          //Dec(FaceIndex2);
          for FaceIndex := FaceIndex2 to F -1 do
          begin
            Faces[FaceIndex-1] := Faces[FaceIndex+1];
          end;
          Dec(F,2);
          Setlength(Faces,F);
          break;
        end;
      end;
    end;
  until not DuplicateFound;

  SetLength(Box,F);

  // initialize NodeUsed array
  SetLength(NodeUsed,V);
  for NodeIndex := 0 to V-1 do
  begin
    NodeUsed[NodeIndex] := False;
  end;

  // check which nodes were used.
  for FaceIndex1 := 0 to F-1 do
  begin
    for FaceIndex2 := 0 to DIM-1 do
    begin
      NodeUsed[Faces[FaceIndex1,FaceIndex2]] := True;
    end;
  end;

  // eliminate nodes that were not used.
  SetLength(NodeIndicies,V);

  EliminatedNodes := 0;
  for NodeIndex := 0 to V-1  do
  begin
    if (EliminatedNodes > 0) {and (NodeIndex+EliminatedNodes < V)} then
    begin
      Vertices[NodeIndex-EliminatedNodes] := Vertices[NodeIndex];
    end;
    if not NodeUsed[NodeIndex] then
    begin
      Inc(EliminatedNodes);
    end;
    NodeIndicies[NodeIndex] := NodeIndex-EliminatedNodes;
  end;
  if EliminatedNodes > 0 then
  begin
    V := V-EliminatedNodes;
    SetLength(Vertices,V);
    for FaceIndex1 := 0 to F-1 do
    begin
      for FaceIndex2 := 0 to DIM-1 do
      begin
        Faces[FaceIndex1,FaceIndex2] := NodeIndicies[Faces[FaceIndex1,FaceIndex2]];
      end;
    end;
  end;

{  for NodeIndex := V-1 downto 0 do
  begin
    if not NodeUsed[NodeIndex] then
    begin
      for InnerNodeIndex := NodeIndex to V-2 do
      begin
        Vertices[InnerNodeIndex] := Vertices[InnerNodeIndex +1];
      end;
      Dec(V);
      SetLength(Vertices,V);
      // update face indicies.
      for FaceIndex1 := 0 to F-1 do
      begin
        for FaceIndex2 := 0 to DIM-1 do
        begin
          if Faces[FaceIndex1,FaceIndex2] > NodeIndex then
          begin
            Dec(Faces[FaceIndex1,FaceIndex2]);
          end;
        end;
      end;
    end;
  end;    }
end;

function TPolyhedron.FacesSame(FaceIndex1, FaceIndex2: Integer): boolean;
var
  Face1, Face2: tPointi;
begin
  Face1 := Faces[FaceIndex1];
  Face2 := Faces[FaceIndex2];
  result := False;
  if Face1[X] = Face2[X] then
  begin
    if Face1[Y] = Face2[Y] then
    begin
      result := (Face1[Z] = Face2[Z]);
    end
    else if Face1[Y] = Face2[Z] then
    begin
      result := (Face1[Z] = Face2[Y]);
    end;
  end
  else if Face1[X] = Face2[Y] then
  begin
    if Face1[Y] = Face2[X] then
    begin
      result := (Face1[Z] = Face2[Z]);
    end
    else if Face1[Y] = Face2[Z] then
    begin
      result := (Face1[Z] = Face2[X]);
    end;
  end
  else if Face1[X] = Face2[Z] then
  begin
    if Face1[Y] = Face2[X] then
    begin
      result := (Face1[Z] = Face2[Y]);
    end
    else if Face1[Y] = Face2[Y] then
    begin
      result := (Face1[Z] = Face2[X]);
    end;
  end
end;

{ TSegmentObject }

function TSegmentObject.Length: extended;
begin
  result := Sqrt(Sqr(StartLoc[X] - EndLoc[X]) + Sqr(StartLoc[Y] - EndLoc[Y])
    + Sqr(StartLoc[Z] - EndLoc[Z]));
end;

procedure SetEpsilon;
begin
  Epsilon := 1;
  while 1 + Epsilon/2 > 1 do
  begin
    Epsilon := Epsilon/2;
  end;
  Epsilon := Epsilon*1e9;
end;

function TSegmentObject.TwoDIntersection(ASegment : TSegmentObject;
      var Intersection : TPointd) : boolean;
var
  Lower1, Higher1, Lower2, Higher2 : extended;
  DimIndex : integer;
  num, denom : double;
  s, t : double;
  a, b, c, d : TPointd;
begin
//  result := True;
  for DimIndex := 0 to 1 do
  begin
    if StartLoc[DimIndex] < EndLoc[DimIndex] then
    begin
      Lower1 := StartLoc[DimIndex];
      Higher1 := EndLoc[DimIndex];
    end
    else
    begin
      Higher1 := StartLoc[DimIndex];
      Lower1 := EndLoc[DimIndex];
    end;
    if ASegment.StartLoc[DimIndex] < ASegment.EndLoc[DimIndex] then
    begin
      Lower2 := ASegment.StartLoc[DimIndex];
      Higher2 := ASegment.EndLoc[DimIndex];
    end
    else
    begin
      Higher2 := ASegment.StartLoc[DimIndex];
      Lower2 := ASegment.EndLoc[DimIndex];
    end;
    result := (Higher1 >= Lower2) and (Higher2 >= Lower1);
    if not result then
    begin
      Exit;
    end;
  end;
  if result then
  begin
    a := StartLoc;
    b := EndLoc;
    c := ASegment.StartLoc;
    d := ASegment.EndLoc;
    denom := a[X] * (d[Y]- c[Y])
           + b[X] * (c[Y]- d[Y])
           + d[X] * (b[Y]- a[Y])
           + c[X] * (a[Y]- b[Y]);
    result := (denom<>0);
    if not result then
    begin
      Exit;
    end;
  end;
  if result then
  begin
    num := a[X] * (d[Y] - c[Y])
         + c[X] * (a[Y] - d[Y])
         + d[X] * (c[Y] - a[Y]);
    s := num/denom;

    num := - (a[X] * (c[Y] - b[Y])
            + b[X] * (a[Y] - c[Y])
            + c[X] * (b[Y] - a[Y]));
    t := num/denom;

    Intersection[X] := a[X] + s * (b[X] - a[X]);
    Intersection[Y] := a[Y] + s * (b[Y] - a[Y]);
    Intersection[Z] := a[Z] + s * (b[Z] - a[Z]);

    result := (0 <= s) and (s <= 1) and (0 <= t) and (t <= 1);

  end;
end;

Initialization
  SetEpsilon;

end.
